library(PtProcess)
library(minpack.lm)

#### Calulation of k ####

#This function f_k will count the number of different values
# of the ACER functions between eta_1 and eta_f
f_k<- function(X,k,eta1, etaf, alpha)
{
  
  # aux <- CI_boot(X,mean(X),k, 50, alpha)
  aux <- CI_Poisson(X, mean(X),k)
  aux1 <- aux$epsilon
  aux_eps <- aux1
  nb_points <- 100
  if(etaf != eta1){
    pas <- abs(etaf - eta1)/(nb_points)
  }
  else{
    pas <- 0
  }
  
  test_level <- seq(eta1, etaf, pas)
  eps <- rep(0,nb_points)
  for(level in test_level)
  {
    # aux <- CI_boot(X, level,k, 100)
    aux <- CI_Poisson(X, level,k)
    
    aux1 <- aux$epsilon
    eps <- c(eps, aux1)
  }

  v_dif <- length(unique(eps[eps!=0]))

  return(v_dif)
}

#This function detect_k will choose a k to 
# get a value of k as small as possible without
# having a lack of precision with long sequences
#of constant values
detect_k <- function(X, eta1, etaf, alpha)
{
  N <- length(X)
  M <- 105
  m <- 5
  bi <- f_k(X,m, eta1, etaf, alpha)

  q <- floor(0.8*bi)
  while((M-m) > 5){
    m_aux <- floor((M+m)/2)
    v <- f_k(X,m_aux, eta1, etaf, alpha)
    if(v < q){
      M <- m_aux
    }
    else{
      m <- m_aux
    }
  }
  return(m)
}



#### Calculation of etaf1, eta_f ####
#eta1 will represent the lowest level and eta_f the 
#highest of our study
eta_detect <- function(X, k, alpha)
{
  nb_values <- 200
  pas <- (max(X)-min(X))/nb_values
  test_level <- seq(min(X),max(X),pas)
  i <- 1
  k_i <- k
  aux_end <- rep(0,k_i)
  for(level in test_level)
  {
    # aux <- CI_boot(X,level,k_i,100, alpha)
    aux <- CI_Poisson(X, level, k_i)
    aux1 <- aux$epsilon
    if(aux1>0){
      aux_end[i] <- (aux1)

    }
    else{
      aux_end[i] <- 0
    }
    i = i+1
  }
  m <- max(aux_end)
  ind <- max(which(aux_end == m))
  v <- aux_end[ind:length(aux_end)]
  mu1 <- test_level[ind]
  w <- diff(log(v))
  l <- which(w != 0)
  r <- diff(l)

  m <- min(which(r >8))
  p <- l[m]
  muf <- test_level[p]
  
  return(c(mu1, muf))
}



#### Detection of exceedances ####
sum_validate_data <- function(X, level)
{
  N <- length(X)
  a <- rep(0, N-1)
  # Calculation of a, number of exceedance respecting 
  # the condition
  index <- which(X > level)
  l <- length(index)
  if(l==0)
  {
    return(a)
  }
  ### firt index
  aux <- index[1] 
  if(aux > 0)
  {
    a[1:(aux)] <- a[1:(aux)] + rep(1,aux)
  }
  if(l >1)
  {
    for (i in 1:(l-1))
    {
      
      aux <- index[i+1] - index[i] - 1
      if(aux >= 1)
      {
        a[1:(aux)] <- a[1:(aux)] + rep(1,aux)
      }
    }
  }
  
  return(a)
}


#### ACER values ####
epsilon_app <- function(X,aux_sum, k)
{
  # Here we just divide the number of exceedance 
  # by approximately the number of points to get a probability
  epsilon <- aux_sum[k]/(length(X)+1-k+1)
  return(epsilon)
  # }
}


#### Confidence intervals ####
CI_Poisson <- function(X, level, k)
{
  #First we calculate the value of epsilon
  x <- sum_validate_data(X, level)
  N <- length(X)
  eps_aux <- epsilon_app(X,x, k)
  
  # Then we apply the classical formula of a CI for a Poisson 
  #distribution
  if(eps_aux == 0){
    C_plus <- 1.96*sqrt((N-k+1))
    C_minus <- -1.96*sqrt((N-k+1))
  }
  else{
    C_plus <- eps_aux*(1 + 1.96/sqrt((N-k+1)*eps_aux))
    C_minus <- eps_aux*(1 - 1.96/sqrt((N-k+1)*eps_aux))
  }
  
  
  C <- data.frame(C_plus, C_minus)
  return(list("CI" = C, "epsilon" = eps_aux))
}

CI_boot <- function(X, level, k, nb, alpha)
{
  nb_sampling <- nb
  L <- rep(0, nb_sampling)
  for(j in 1:nb_sampling){
    # We calculate the value of epsilon for sample based on 
    # the initial data set.
    x <- sum_validate_data(sample(X, replace = T), level)
    L[j] <- epsilon_app(X,x, k)
  }
  
  # we take the alpha % confidence interval
  aux <- sort(L)
  C_minus <- aux[floor(alpha/2*length(aux))]
  C_plus <- aux[floor((1-alpha/2)*length(aux))+1]
  
  C <- list(C_plus, C_minus)
  x <- sum_validate_data(X, level)
  eps_aux <- epsilon_app(X,x, k)

  return(list("CI" = C, "epsilon" = eps_aux))
}



#### ACER method ####
ACER_General <- function(X, alpha)
{
  #First we fixe k at a possible value
  # and we find eta1 and eta_f. These two values 
  #do not vary too much with k
  eta <- eta_detect(X, 2, alpha)
  eta1 <- eta[1]
  etaf <- eta[2]

  #With eta1 and eta_f, we find k.
  k <- detect_k(X, eta1, etaf, alpha)
  
  #We calculate then the true values of eta1 and eta_f
  eta <- eta_detect(X, k, alpha)
  eta1 <- eta[1]
  etaf <- eta[2]
  
  nb_points <- 200
  
  pas <- (etaf-eta1)/(nb_points)
  test_level <- seq(eta1, etaf, pas)
  aux_eps <- 0
  aux_w <- 0
  for(level in test_level)
  {
    # aux <- CI_boot(X, level,k, 100)
    aux <- CI_Poisson(X, level, k)
    aux1 <- aux$epsilon
    aux2 <- unlist(aux$CI)
    v <- 1/(log(abs(aux2[1])) -log(abs(aux2[2])))^2
    
    aux_eps <- c(aux_eps, aux1)
    aux_w <- c(aux_w, v)
    
  }
  
  eps <- aux_eps[2:length(aux_eps)]
  w <- aux_w[2:length(aux_w)]
  
  level <- test_level
  
  y <- log(eps)
  # plot(w)
  y_m <- sum(w*y) / sum(w)
  
  ACER_sim_vect <- function(param){
    #we are creating now the function for the given k
    b <- param[2]
    c <- param[3]
    a_t <- param[1]
    x <- log(abs(1 + a_t*abs(level - b)^c))
    x_m <- sum(w*x)/sum(w)
    gam <- -(sum(w*(x-x_m)*(y-y_m))/sum((w*(x-x_m)^2)))
    q <- exp(y_m + gam*x_m)
    result <- rep(0, length(w))
    for(i in 1:length(w)){
      result[i] <- as.double(sqrt(w[i])*abs(y[i] - log(q) 
               + gam*log(abs(1 + a_t*abs(level[i] - b)^c))))
    }
    return(unlist(result))
  }
  #
  #
  m <- min(X)
  # ## we need to find b, c
  minit <- (m+eta1)/2
  
  
  o <- nls.lm(c(1,minit, 1), lower = c(0,m, 0), 
        upper = c(Inf,eta1, 5), ACER_sim_vect,  jac = NULL, 
        nls.lm.control(ftol = 0.001,maxiter = 1000)) 
  
  
  param <- o$par
  ##Now, we need to find a, q_ln
  a_t <- param[1]
  b <- param[2]
  c <- param[3]
  x <- log((1 + a_t*abs(level - b)^c))
  x_m <- sum(w*x)/sum(w)

  gam <- -(sum(w*(x-x_m)*(y-y_m))/sum((w*(x-x_m)^2)))

  q <- exp((y_m + gam*x_m))
  
  
  # print(c(a_t,b,c,q,gam))
  
  lres <- rep(0, length(eps))
  
  for(l in 1:length(eps)){
    lres[l] <- log(q) - gam*log(1+a_t*abs(level[l] - b)^c)
  }

  return(list(coef = c(a_t, b, c, q, gam), eps = eps, 
              level = level))

}



#### Calculation VaR ####
VaR <- function(X, alpha){
  v <- ACER_General(X, alpha)
  coef <- v$coef
  coef <- unlist(coef)
  a_t <- coef[1]
  b <- coef[2]
  c <- coef[3]
  q <- coef[4]
  gam <- coef[5]
  # VaR <- b+abs(1/a_t*abs(alpha/q)^(1/gam) - 1)^(1/c)
  VaR_new <- b + abs(((alpha/q)^(-1/gam)-1)/a_t)^(1/c)
  while(VaR_new > max(X))
  {
    VaR_new <- VaR(X,alpha)
  }
  return(VaR_new)
}